//@ts-check
"use strict";

/**
@typedef SoundRes
@type {{
  sound: PIXI.sound.Sound,
  name: string;
  referenceVolume: number,
}}
 */

/**
@typedef AudioFade
@type {{
  kind: "in " | "out";
  started: Date;
  soundRes: SoundRes,
  cancel: () => void;
}}
 */

/** @type {Map<SoundRes, AudioFade>} */
const fades = new Map();

/** @type {Map<string, SoundRes>} */
const cachedSounds = new Map();

const inGameFadeDuration = 1.5;

function registerAudioResources() {
  app.loader.add([
    "music/music.json",
    "sounds/sounds.json",
    "sounds/silence.wav"
  ]);
}

function registerAudioSounds() {
  const bindings = app.loader.resources[`sounds/sounds.json`].data;

  function recurse(obj) {
    if (Array.isArray(obj)) {
      const path = `sounds/${obj[0]}`;
      if (app.loader.resources[path] == null) app.loader.add(path);
    } else if (typeof obj === "string") {
      const path = `sounds/${obj}`;
      if (app.loader.resources[path] == null) app.loader.add(path);
    } else {
      for (const key in obj) recurse(obj[key]);
    }
  }

  recurse(bindings);
}

/**
 * @param {string} path
 * @param {number} referenceVolume
 * @returns {SoundRes}
 */
function getMusicSound(path, referenceVolume) {
  if (!referenceVolume) {
    referenceVolume = 1.0;
  }
  if (location.search.includes("nomusic")) {
    referenceVolume = 0;
  }

  const cached = cachedSounds.get(path);
  if (typeof cached !== "undefined") {
    return cached;
  }

  let sound = PIXI.sound.Sound.from(`assets/music/${path}`);
  sound.singleInstance = true;
  let tokens = path.split(`/`);
  /** @type {SoundRes} */
  let res = {
    name: tokens[tokens.length - 1].replace(/ /g, "-"),
    referenceVolume,
    sound,
  };
  cachedSounds.set(path, res);
  return res;
}

let globalBgmCue = null;
/** @type {SoundRes} */
let globalBgmSound = null;
/** @type {SoundRes} */
let combatBgmSound = null;
let globalBgmVolume = 1;
let isInCombat = false;

if (location.search.includes("mute")) PIXI.sound.volumeAll = 0;
else PIXI.sound.volumeAll = 0.5;

/** @param {string} cue */
function ensureMenuMusic(cue) {
  if (cue === globalBgmCue) return;
  if (globalBgmSound != null) fadeOutSound(globalBgmSound, 0.5, "stop");
  if (combatBgmSound != null) fadeOutSound(combatBgmSound, 0.5, "stop");
  globalBgmSound = null;
  combatBgmSound = null;

  globalBgmCue = cue;
  const binding = traverse(app.loader.resources["music/music.json"].data, cue);

  if (binding == null) {
    debug(`Missing music binding for cue: ${cue}`);
    return;
  }

  globalBgmSound = getMusicSound(binding.path, binding.volume);
  globalBgmSound.sound.play({ loop: true });
  fadeInSound(globalBgmSound, inGameFadeDuration);
}

function ensureInGameMusic(cue) {
  if (cue === globalBgmCue) return;
  console.log(`---------------------------------------`);
  console.log(`Ensuring in-game music ${cue}`);
  if (globalBgmSound != null) fadeOutSound(globalBgmSound, inGameFadeDuration, "pause");
  if (combatBgmSound != null) fadeOutSound(combatBgmSound, inGameFadeDuration, "pause");
  globalBgmSound = null;
  combatBgmSound = null;

  globalBgmCue = cue;
  const binding = traverse(app.loader.resources["music/music.json"].data, cue);

  if (binding == null) {
    debug(`Missing music binding for cue: ${cue}`);
    return;
  }

  const { base, fight, volume } = binding;
  globalBgmVolume = volume;

  globalBgmSound = getMusicSound(base, volume);
  if (globalBgmSound.sound.paused) {
    globalBgmSound.sound.resume();
  } else if (!globalBgmSound.sound.isPlaying) {
    globalBgmSound.sound.play({ loop: true });
  }
  fadeInSound(globalBgmSound, inGameFadeDuration);

  if (fight != null) {
    combatBgmSound = getMusicSound(fight, volume);
    if (combatBgmSound.sound.paused) {
      combatBgmSound.sound.resume();
    } else if (!combatBgmSound.sound.isPlaying) {
      combatBgmSound.sound.play({ loop: true });
    }
    if (isInCombat) {
      fadeInSound(combatBgmSound, inGameFadeDuration);
    } else {
      combatBgmSound.sound.volume = 0;
    }
  }
}

function updateCombatBgmVolume() {
  if (combatBgmSound != null) {
    if (isInCombat) {
      if (combatBgmSound.sound.isPlaying) {
        fadeInSound(combatBgmSound, inGameFadeDuration);
      }
    } else {
      fadeOutSound(combatBgmSound, inGameFadeDuration, "continue");
    }
  }
}

/**
@typedef AfterFadeOut
@type {"continue" | "stop" | "pause"}
 */

/**
 * @param {SoundRes} soundRes
 * @param {number} duration
 * @param {AfterFadeOut} after
 */
function fadeOutSound(soundRes, duration, after) {
  let state = {
    cancelled: false,
  };
  /** @type {AudioFade} */
  let fade = {
    soundRes,
    kind: "out",
    started: new Date(),
    cancel: () => { state.cancelled = true; }
  };
  setAudioFade(soundRes, fade);
  let originalVolume = soundRes.sound.volume;

  duration = duration || 0.2;
  let timer = 0;

  function fadeOut() {
    if (state.cancelled) {
      return app.ticker.remove(fadeOut);
    }

    if (!soundRes.sound.isPlaying) {
      clearAudioFade(soundRes, fade, "ended");
      app.ticker.remove(fadeOut);
      return;
    }

    const dt = app.ticker.deltaMS / 1000;
    timer = Math.min(timer + dt, duration);
    soundRes.sound.volume = originalVolume * (1 - timer / duration);
    if (timer === duration) {
      clearAudioFade(soundRes, fade, "ended");
      soundRes.sound.volume = 0;
      app.ticker.remove(fadeOut);
      switch (after) {
        case "pause":
          console.log(`>> after fade: pausing | `, soundRes.name);
          soundRes.sound.pause();
          break;
        case "stop":
          console.log(`>> after fade: stopping | `, soundRes.name);
          soundRes.sound.stop();
          break;
        default:
          console.log(`>> after fade: continuing in bg |`, soundRes.name);
          break;
      }
    }
  }

  app.ticker.add(fadeOut);
}

/**
 * @param {SoundRes} sound
 * @param {AudioFade} fade
 */
function setAudioFade(sound, fade) {
  let prev = fades.get(sound);
  if (prev) {
    console.warn(`fade-` + prev.kind, prev.soundRes.name, `CANCELLING`);
    prev.cancel();
  }
  console.log(`fade-` + fade.kind, fade.soundRes.name, `STARTING`);
  fades.set(sound, fade);
}

/**
 * @param {SoundRes} sound
 * @param {AudioFade} fade
 * @param {"ended" | "cancelled"} reason
 */
function clearAudioFade(sound, fade, reason) {
  if (fades.get(sound) === fade) {
    console.log(`fade-` + fade.kind, fade.soundRes.name, `CLEARING`, `(${reason})`);
    fades.delete(sound);
  }
}

/** @param {SoundRes} soundRes @param {number=} duration */
function fadeInSound(soundRes, duration) {
  let state = {
    cancelled: false,
  };
  /** @type {AudioFade} */
  let fade = {
    soundRes,
    kind: "in ",
    started: new Date(),
    cancel: () => { state.cancelled = true; }
  };
  setAudioFade(soundRes, fade);

  duration = duration || 0.2;
  let timer = 0;

  function fadeIn() {
    if (state.cancelled) {
      return app.ticker.remove(fadeIn);
    }

    const dt = app.ticker.deltaMS / 1000;
    timer = Math.min(timer + dt, duration);
    if (!soundRes.sound.isPlaying || timer === duration) {
      clearAudioFade(soundRes, fade, "ended");
      soundRes.sound.volume = soundRes.referenceVolume;
      return app.ticker.remove(fadeIn);
    }

    soundRes.sound.volume = soundRes.referenceVolume * (timer / duration);
  }

  app.ticker.add(fadeIn);
}

/**
 * @param {string} cue
 * @returns {PIXI.sound.Sound}
 * */

function getSound(cue) {
  let { path, volume } = traverse(app.loader.resources["sounds/sounds.json"].data, cue);

  if (path == null) {
    debug(`Missing sound binding for cue: ${cue}`);
    path = "silence.wav";
  }

  const sound = app.loader.resources[`sounds/${path}`].sound;
  sound.volume = volume;
  return sound;
}

/**
 * @param {string} cue
 * @param {PIXI.Point} position
 * */
function play2Dsound(cue, position) {
  const dist = distxy(inGame.baxton.pos, position);
  if (dist > 10) return;

  const sound = getSound(cue);
  sound.volume *= 1 - dist / 10;
  sound.play();
}
